from jaikit.character.unicode import WIKI_UNICODE_INFO
from itertools import chain


# 包含汉字的区段
CHN_PART = WIKI_UNICODE_INFO.loc[WIKI_UNICODE_INFO["文字"].str.contains("汉字", na=False)]
CHINESE_ORDS = {
    _
    for _ in chain(
        *(range(item["start"], item["end"] + 1) for idx, item in CHN_PART.iterrows())
    )
}
REMOVE_CHINESE_MAP = {_: "" for _ in CHINESE_ORDS}


def contains_chinese(text: str) -> bool:
    """基于Unicode字符集的汉字所涉及范围，判断一个字符串是否含有汉字"""
    return any((ord(character) in CHINESE_ORDS for character in text))


def remove_chinese(text: str) -> str:
    """基于Unicode码表汉字区段，从字符串中删除所有汉字"""
    return text.translate(REMOVE_CHINESE_MAP)



from tensorflow.keras.callbacks import History
from typing import List
from matplotlib import pyplot as plt


def plot_metric(history: History, metric: str, save_path: str = ""):
    """
    将train和validation的history数据绘制出来
    history.history中key为"loss"、"auc"的list存储的是训练集上的指标
    而key为"val_loss"和"val_auc"的list存储的是验证集上的指标
    """
    fig = plt.figure()
    train_metrics: List[float] = history.history[metric]
    val_metrics: List[float] = history.history.get("val_" + metric)
    epochs = range(1, len(train_metrics) + 1)
    plt.plot(epochs, train_metrics, "bo--")  # 蓝色描点虚线连接
    title = "Training"
    legend = ["train_" + metric]
    if val_metrics:
        plt.plot(epochs, val_metrics, "ro-")  # 红色描点实线连接
        title += " and validation "
        legend.append("val_" + metric)
    plt.title(f"{title} {metric}")
    plt.xlabel("Epochs")
    plt.ylabel(metric)
    plt.legend(legend)  # list的元素与前面多条plot语句所绘制内容对应
    plt.show()
    if save_path:
        fig.savefig(save_path)
        print(f"图像已保存在: {save_path}")



"""
将平铺的（child, parent）形式的数据转为树状结构，借此进行层次数据呈现与后续处理

"""
from collections import deque
from typing import List


class Node:
    def __init__(self, value):
        self.value = value
        self.children = []

    def add_child(self, obj):
        self.children.append(obj)


class FlatRelationToTree:
    def __init__(self, relations: List[dict]):
        self.relations = relations
        self.all_nodes = {rel["_id"]: Node(value=rel) for rel in self.relations}
        for rel in self.relations:
            if rel["parent_id"] is not None:
                self.all_nodes[rel["parent_id"]].add_child(self.all_nodes[rel["_id"]])
        self.root_nodes = [
            node
            for idx, node in self.all_nodes.items()
            if node.value["parent_id"] is None
        ]
        self.id_to_parent = dict()
        for rel in self.relations:
            self.id_to_parent.update({rel["_id"]: rel["parent_id"]})

    def find_all_ancestors(self, node_id: int):
        parent_id = node_id
        ids_order = deque()
        while parent_id is not None:
            ids_order.appendleft(parent_id)
            parent_id = self.id_to_parent[parent_id]
        return "-".join([self.all_nodes[idx].value["name"] for idx in ids_order])

    def all_whole_chains(self):
        whole_cont = []
        for rel in self.relations:
            whole_cont.append(self.find_all_ancestors(rel["_id"]))
        return sorted(whole_cont)

    def id_to_whole_chain(self):
        return {idx: self.find_all_ancestors(node_id=idx) for idx in self.id_to_parent}



import unicodedata


CJK_Char_LIST = [
    [range(13312, 19893), "CJK Unified Ideographs Extension A 3.0"],
    [range(19968, 40869), "CJK Unified Ideographs 1.1"],
    [range(40870, 40891), "CJK Unified Ideographs 4.1"],
    [range(40892, 40899), "CJK Unified Ideographs 5.1"],
    [range(40900, 40907), "CJK Unified Ideographs 5.2"],
    [range(40908, 40909), "CJK Unified Ideographs 6.1"],
    [range(40909, 40917), "CJK Unified Ideographs 8.0"],
    [range(40918, 40938), "CJK Unified Ideographs 10.0"],
    [range(40939, 40943), "CJK Unified Ideographs 11.0"],
    [range(63744, 64045), "CJK Compatibility Ideographs 1.1"],
    [range(64046, 64047), "CJK Compatibility Ideographs 6.1"],
    [range(64048, 64106), "CJK Compatibility Ideographs 3.2"],
    [range(64107, 64109), "CJK Compatibility Ideographs 5.2"],
    [range(64112, 64217), "CJK Compatibility Ideographs 4.1"],
    [range(131072, 173782), "CJK Unified Ideographs Extension B 3.1"],
    [range(173824, 177972), "CJK Unified Ideographs Extension C 5.2"],
    [range(177984, 178205), "CJK Unified Ideographs Extension D 6.0"],
    [range(178208, 183983), "CJK Unified Ideographs Extension E 8.0"],
    [range(183984, 191456), "CJK Unified Ideographs Extension F 10.0"],
    [range(194560, 195101), "CJK Compatibility Supplement 3.1"],
    [range(11904, 11929), "CJK Radicals Supplement 3.0"],
    [range(11931, 12019), "CJK Radicals Supplement 3.0"],
    [range(12032, 12245), "Kangxi Radicals 3.0"],
    [range(12272, 12283), "Ideographic Description Characters 3.0"],
    [range(12288, 12343), "CJK Symbols and Punctuation 1.1"],
    [range(12344, 12346), "CJK Symbols and Punctuation 3.0"],
    [range(12347, 12349), "CJK Symbols and Punctuation 3.2"],
    [range(12350, 12351), "CJK Symbols and Punctuation 3.0"],
    [range(12351, 12352), "CJK Symbols and Punctuation 1.1"],
    [range(12549, 12588), "Bopomofo 1.1"],
    [range(12589, 12590), "Bopomofo 5.1"],
    [range(12688, 12703), "Kanbun 1.1"],
    [range(12704, 12727), "Bopomofo Extended 3.0"],
    [range(12736, 12751), "CJK Strokes 4.1"],
    [range(12752, 12771), "CJK Strokes 5.1"],
    [range(12832, 12867), "Enclosed CJK Letters and Months 1.1"],
    [range(12928, 12976), "Enclosed CJK Letters and Months 1.1"],
    [range(12992, 13003), "Enclosed CJK Letters and Months 1.1"],
    [range(13144, 13168), "CJK Compatibility 1.1"],
    [range(13179, 13183), "CJK Compatibility 1.1"],
    [range(13280, 13310), "CJK Compatibility 1.1"],
]


def CJK_detail(Character):
    try:
        Udata = unicodedata.name(Character)
    except:
        Udata = ""
    for _ in CJK_Char_LIST:
        if ord(Character) in _[0]:
            return ord(Character), _[1], Udata
    return ord(Character), "Not CJK", Udata



from typing import Dict

from jaikit.datastructure.trie import Trie


class MaxMatchTranslate:
    """
    按照正向最大匹配的原则，将文本中的词对照translate_map映射表进行翻译替换
    """
    def __init__(self, translate_map: Dict[str, str]):
        self.translate_map = translate_map
        self.max_len = max([len(item) for item in self.translate_map])
        self.trie = Trie()
        for word in self.translate_map:
            self.trie.insert(word)

    def max_match_translate(self, text: str):
        start = 0
        result = ""
        while start < len(text):
            match_window_len = 0
            match_word = ""
            for cur in range(1, self.max_len + 1):
                if start + cur > len(text):
                    break
                window = text[start : start + cur]
                if self.trie.search(window):
                    match_window_len = cur
                    match_word = window
            if match_window_len > 0:
                start += match_window_len
                result += self.translate_map[match_word]
            else:
                start += 1
                result += text[start - 1]
        return result



import pickle
from itertools import permutations
from typing import List, Tuple, Optional


def del_sub(pos_ls: List[Tuple]) -> List[Tuple]:
    """
    一个范围是另一范围的子集时，删除子集范围
    :pos_ls:原tuple列表
    :return:删除子集后的tuple列表
    """
    del_set = set()  # 要删除的子集
    for tp in list(permutations(list(range(len(pos_ls))), 2)):
        if (
            pos_ls[tp[0]][1] <= pos_ls[tp[1]][1]
            and pos_ls[tp[0]][0] >= pos_ls[tp[1]][0]
        ):  # 子集判断
            del_set.add(tp[0])  # 如果是子集，就添加
    return [_ for _idx, _ in enumerate(pos_ls) if not _idx in del_set]


class Trie:
    def __init__(self, load_path: Optional[str]=None):
        if load_path is None:
            self.root = {}
        else:
            with open(load_path, "rb") as f:
                self.root = pickle.load(f)
        self.word_end = -1

    def insert(self, word):
        curNode = self.root
        for c in word:
            if not c in curNode:
                curNode[c] = {}
            curNode = curNode[c]
        curNode[self.word_end] = True

    def search(self, word) -> bool:
        curNode = self.root
        for c in word:
            if not c in curNode:
                return False
            curNode = curNode[c]
        if self.word_end not in curNode:
            return False
        return True

    def starts_with(self, prefix) -> bool:
        curNode = self.root
        for c in prefix:
            if not c in curNode:
                return False
            curNode = curNode[c]
        return True

    def scan_item(self, target, greedy_mode=False):
        # TODO 从target中找出命中的，需要改 t.scan_item("aba") ['ab', 'ba']重复
        if not greedy_mode:
            pos_ls = []
            found_ls = []
            for _idx, _ in enumerate(target):
                if self.starts_with(_):
                    for id in range(len(target) - _idx + 1):
                        if self.search(target[_idx : _idx + id]):
                            pos_ls.append((_idx, _idx + id))
            for pos_tp in del_sub(pos_ls):
                fd = target[pos_tp[0] : pos_tp[1]]
                found_ls.append(fd)
        else:
            found_ls = []
            for _idx, _ in enumerate(target):
                if self.starts_with(_):
                    for id in range(len(target) - _idx + 1):
                        fd = target[_idx : _idx + id]
                        if self.search(fd):
                            found_ls.append(fd)
        return found_ls

    def remove_item(self, word: str) -> dict:
        """从Trie树中删除某词"""
        if not self.search(word):
            return self.root
        leaf_dict = self.root
        upper_dict = leaf_dict
        for char in word[:-1]:
            upper_dict = leaf_dict
            leaf_dict = leaf_dict[char]
        if len(leaf_dict[word[-1]]) == 1:
            leaf_dict.pop(word[-1])
            upper_dict.pop(word[-2], None)
        else:
            leaf_dict[word[-1]].pop(-1)
        return self.root



import re

import pandas as pd
from pandas import Series, DataFrame


def series_onehot_converter(series: Series, dimension: int) -> DataFrame:
    """
    将DataFrame的一个整数值列（即Series）转为N维的one-hot表示，返回N列的DataFrame
    这是为了增强pandas.get_dummies的功能，因为get_dummies只能按照所有出现过的值种类数作为one-hot的dimension，不能自由设定
    注意必须符合以下两个条件，否则会报错：
    （1）参数dimension要大于等于series中的最大值
    （2）series中的数值类型必须为整数型
    """
    if dimension < series.max():
        raise ValueError(
            "Dimension must be more than or equal to the maximum value in the series."
        )
    if not re.match(r"int|uint", series.dtypes.name):
        raise TypeError("Values in the series must be in the type of int.")
    return pd.get_dummies(
        series.append(pd.DataFrame(data=range(dimension))[0], ignore_index=True,)
    )[:-dimension]



"""
Unicode字符集工具箱
用途：
查询某字符的Unicode所属区段及相关信息

示例：

    # >>> from jaikit.character.unicode import character_unicode_info
    # >>> character_unicode_info(character="我")

    {'平面': '0 BMP',
     '区段范围': 'U+4E00..U+9FFF',
     '区段名称': '中日韩统一表意文字',
     '英文名称': 'CJK Unified Ideographs',
     '码位数': '20,992',
     '已定义字元数': '20,989',
     '文字': '汉字',
     'start': 19968,
     'end': 40959}

参考资料：
维基百科页面2021年9月3日版本
https://zh.wikipedia.org/wiki/Unicode区段
https://liyucang-git.github.io/2019/06/17/彻底弄懂Unicode编码/
"""
import bisect

import pandas as pd

WIKI_UNICODE_INFO = pd.read_pickle("jaikit/character/unicode_info.pkl")

breakpoints = WIKI_UNICODE_INFO["start"]
section_indexes = list(range(-1, len(WIKI_UNICODE_INFO["start"])))


def character_unicode_info(character: str):
    unicode_ord: int = ord(character)
    info = WIKI_UNICODE_INFO.iloc[section_indexes[bisect.bisect(breakpoints, unicode_ord)]]
    if pd.isna(info["平面"]):
        raise ValueError("不是Unicode字符")
    else:
        info = info.append(pd.Series({"Unicode编码": unicode_ord}))
        return info



"""The setup script."""

from setuptools import find_packages, setup

with open("README.md") as readme_file:
    readme = readme_file.read()

install_requires = ["numpy", "pandas", "tensorflow", "matplotlib"]

setup_requirements = []
develop_requirements = ["ipython", "unittest", "ipdb", "black", "pre-commit", "watchdog", "sphinx", "coverage",
                        "sphinx-rtd-theme"]
test_requirements = []

setup(
    author="Jeffrey Ma",
    author_email="mjw814@163.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    description="Jingwei AI Kit for AI exploration and development.",
    install_requires=install_requires,
    long_description=readme,
    include_package_data=True,
    keywords="jaikit",
    name="jaikit",
    packages=find_packages(include=["jaikit*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/DoliteMatheo/JAIKit.git",
    version="0.0.3",
    zip_safe=False,
)



from typing import List

import numpy as np
import math


def calc_cos_sim(vector_a: np.ndarray, vector_b: np.ndarray) -> float:
    """计算两个列向量之间的余弦相似度"""
    num = np.sum(vector_a * vector_b)
    denom = np.linalg.norm(vector_a) * np.linalg.norm(vector_b)
    if denom == 0:  # 任何一个向量的模为0时直接返回0，防止返回Nan
        return 0.0
    cos = num / denom
    return cos


def calc_idf(word: str, all_docs: List[str]) -> float:
    """计算某个word相对于多篇文章的IDF值"""
    idf = math.log(len(all_docs) / (1 + len([doc for doc in all_docs if word in doc])))
    return idf

